@@ -1,6 +1,6 @@
 # -*- coding: utf-8 -*-
 
-# ########################## Copyrights and license ############################
+############################ Copyrights and license ############################
 #                                                                              #
 # Copyright 2012 Michael Stead <michael.stead@gmail.com>                       #
 # Copyright 2012 Vincent Jacques <vincent@vincent-jacques.net>                 #
@@ -8,9 +8,24 @@
 # Copyright 2013 AKFish <akfish@gmail.com>                                     #
 # Copyright 2013 Vincent Jacques <vincent@vincent-jacques.net>                 #
 # Copyright 2013 martinqt <m.ki2@laposte.net>                                  #
+# Copyright 2014 Vincent Jacques <vincent@vincent-jacques.net>                 #
+# Copyright 2016 @tmshn <tmshn@r.recruit.co.jp>                                #
+# Copyright 2016 Jannis Gebauer <ja.geb@me.com>                                #
+# Copyright 2016 Peter Buckley <dx-pbuckley@users.noreply.github.com>          #
+# Copyright 2017 Aaron Levine <allevin@sandia.gov>                             #
+# Copyright 2017 Simon <spam@esemi.ru>                                         #
+# Copyright 2018 Ben Yohay <ben@lightricks.com>                                #
+# Copyright 2018 Gilad Shefer <gshefer@redhat.com>                             #
+# Copyright 2018 Martin Monperrus <monperrus@users.noreply.github.com>         #
+# Copyright 2018 Matt Babineau <9685860+babineaum@users.noreply.github.com>    #
+# Copyright 2018 Shinichi TAMURA <shnch.tmr@gmail.com>                         #
+# Copyright 2018 Steve Kowalik <steven@wedontsleep.org>                        #
+# Copyright 2018 Thibault Jamet <tjamet@users.noreply.github.com>              #
+# Copyright 2018 per1234 <accounts@perglass.com>                               #
+# Copyright 2018 sfdye <tsfdye@gmail.com>                                      #
 #                                                                              #
 # This file is part of PyGithub.                                               #
-# http://pygithub.github.io/PyGithub/v1/index.html                             #
+# http://pygithub.readthedocs.io/                                              #
 #                                                                              #
 # PyGithub is free software: you can redistribute it and/or modify it under    #
 # the terms of the GNU Lesser General Public License as published by the Free  #
@@ -25,8 +40,9 @@
 # You should have received a copy of the GNU Lesser General Public License     #
 # along with PyGithub. If not, see <http://www.gnu.org/licenses/>.             #
 #                                                                              #
-# ##############################################################################
+################################################################################
 
+import urllib
 import github.GithubObject
 import github.PaginatedList
 
@@ -37,6 +53,7 @@ import github.PullRequestComment
 import github.File
 import github.IssueComment
 import github.Commit
+import github.PullRequestReview
 
 
 class PullRequest(github.GithubObject.CompletableGithubObject):
@@ -44,6 +61,9 @@ class PullRequest(github.GithubObject.CompletableGithubObject):
     This class represents PullRequests. The reference can be found here http://developer.github.com/v3/pulls/
     """
 
+    def __repr__(self):
+        return self.get__repr__({"number": self._number.value, "title": self._title.value})
+
     @property
     def additions(self):
         """
@@ -60,6 +80,14 @@ class PullRequest(github.GithubObject.CompletableGithubObject):
         self._completeIfNotSet(self._assignee)
         return self._assignee.value
 
+    @property
+    def assignees(self):
+        """
+        :type: list of :class:`github.NamedUser.NamedUser`
+        """
+        self._completeIfNotSet(self._assignees)
+        return self._assignees.value
+
     @property
     def base(self):
         """
@@ -180,6 +208,14 @@ class PullRequest(github.GithubObject.CompletableGithubObject):
         self._completeIfNotSet(self._issue_url)
         return self._issue_url.value
 
+    @property
+    def labels(self):
+        """
+        :type: list of :class:`github.Label.Label`
+        """
+        self._completeIfNotSet(self._labels)
+        return self._labels.value
+
     @property
     def merge_commit_sha(self):
         """
@@ -316,6 +352,17 @@ class PullRequest(github.GithubObject.CompletableGithubObject):
         self._completeIfNotSet(self._user)
         return self._user.value
 
+    def as_issue(self):
+        """
+        :calls: `GET /repos/:owner/:repo/issues/:number <http://developer.github.com/v3/issues>`_
+        :rtype: :class:`github.Issue.Issue`
+        """
+        headers, data = self._requester.requestJsonAndCheck(
+            "GET",
+            self.issue_url
+        )
+        return github.Issue.Issue(self._requester, headers, data, completed=True)
+
     def create_comment(self, body, commit_id, path, position):
         """
         :calls: `POST /repos/:owner/:repo/pulls/:number/comments <http://developer.github.com/v3/pulls/comments>`_
@@ -365,22 +412,91 @@ class PullRequest(github.GithubObject.CompletableGithubObject):
         }
         headers, data = self._requester.requestJsonAndCheck(
             "POST",
-            self._parentUrl(self._parentUrl(self.url)) + "/issues/" + str(self.number) + "/comments",
+            self.issue_url + "/comments",
             input=post_parameters
         )
         return github.IssueComment.IssueComment(self._requester, headers, data, completed=True)
 
-    def edit(self, title=github.GithubObject.NotSet, body=github.GithubObject.NotSet, state=github.GithubObject.NotSet):
+    def create_review(self, commit, body, event=github.GithubObject.NotSet, comments=github.GithubObject.NotSet):
+        """
+        :calls: `POST /repos/:owner/:repo/pulls/:number/reviews <https://developer.github.com/v3/pulls/reviews/>`_
+        :param commit: github.Commit.Commit
+        :param body: string
+        :param event: string
+        :param comments: list
+        :rtype: :class:`github.PullRequestReview.PullRequestReview`
+        """
+        assert isinstance(commit, github.Commit.Commit), commit
+        assert isinstance(body, str), body
+        assert event is github.GithubObject.NotSet or isinstance(event, str), event
+        assert comments is github.GithubObject.NotSet or isinstance(comments, list), comments
+        post_parameters = {'commit_id': commit.sha, 'body': body}
+        post_parameters['event'] = 'COMMENT' if event == github.GithubObject.NotSet else event
+        if comments is github.GithubObject.NotSet:
+            post_parameters['comments'] = []
+        else:
+            post_parameters['comments'] = comments
+        headers, data = self._requester.requestJsonAndCheck(
+            "POST",
+            self.url + "/reviews",
+            input=post_parameters
+        )
+        self._useAttributes(data)
+        return github.PullRequestReview.PullRequestReview(self._requester, headers, data, completed=True)
+
+    def create_review_request(self, reviewers=github.GithubObject.NotSet, team_reviewers=github.GithubObject.NotSet):
+        """
+        :calls: `POST /repos/:owner/:repo/pulls/:number/requested_reviewers <https://developer.github.com/v3/pulls/review_requests/>`_
+        :param reviewers: list of strings
+        :param team_reviewers: list of strings
+        :rtype: None
+        """
+        post_parameters = dict()
+        if reviewers is not github.GithubObject.NotSet:
+            assert all(isinstance(element, (str, unicode)) for element in reviewers), reviewers
+            post_parameters["reviewers"] = reviewers
+        if team_reviewers is not github.GithubObject.NotSet:
+            assert all(isinstance(element, (str, unicode)) for element in team_reviewers), team_reviewers
+            post_parameters["team_reviewers"] = team_reviewers
+        headers, data = self._requester.requestJsonAndCheck(
+            "POST",
+            self.url + "/requested_reviewers",
+            input=post_parameters
+        )
+
+    def delete_review_request(self, reviewers=github.GithubObject.NotSet, team_reviewers=github.GithubObject.NotSet):
+        """
+        :calls: `DELETE /repos/:owner/:repo/pulls/:number/requested_reviewers <https://developer.github.com/v3/pulls/review_requests/>`_
+        :param reviewers: list of strings
+        :param team_reviewers: list of strings
+        :rtype: None
+        """
+        post_parameters = dict()
+        if reviewers is not github.GithubObject.NotSet:
+            assert all(isinstance(element, (str, unicode)) for element in reviewers), reviewers
+            post_parameters["reviewers"] = reviewers
+        if team_reviewers is not github.GithubObject.NotSet:
+            assert all(isinstance(element, (str, unicode)) for element in team_reviewers), team_reviewers
+            post_parameters["team_reviewers"] = team_reviewers
+        headers, data = self._requester.requestJsonAndCheck(
+            "DELETE",
+            self.url + "/requested_reviewers",
+            input=post_parameters
+        )
+
+    def edit(self, title=github.GithubObject.NotSet, body=github.GithubObject.NotSet, state=github.GithubObject.NotSet, base=github.GithubObject.NotSet):
         """
         :calls: `PATCH /repos/:owner/:repo/pulls/:number <http://developer.github.com/v3/pulls>`_
         :param title: string
         :param body: string
         :param state: string
+        :param base: string
         :rtype: None
         """
         assert title is github.GithubObject.NotSet or isinstance(title, (str, unicode)), title
         assert body is github.GithubObject.NotSet or isinstance(body, (str, unicode)), body
         assert state is github.GithubObject.NotSet or isinstance(state, (str, unicode)), state
+        assert base is github.GithubObject.NotSet or isinstance(base, (str, unicode)), base
         post_parameters = dict()
         if title is not github.GithubObject.NotSet:
             post_parameters["title"] = title
@@ -388,6 +504,8 @@ class PullRequest(github.GithubObject.CompletableGithubObject):
             post_parameters["body"] = body
         if state is not github.GithubObject.NotSet:
             post_parameters["state"] = state
+        if base is not github.GithubObject.NotSet:
+            post_parameters["base"] = base
         headers, data = self._requester.requestJsonAndCheck(
             "PATCH",
             self.url,
@@ -418,20 +536,41 @@ class PullRequest(github.GithubObject.CompletableGithubObject):
 
     def get_comments(self):
         """
+        Warning: this only returns review comments. For normal conversation comments, use get_issue_comments.
+
         :calls: `GET /repos/:owner/:repo/pulls/:number/comments <http://developer.github.com/v3/pulls/comments>`_
         :rtype: :class:`github.PaginatedList.PaginatedList` of :class:`github.PullRequestComment.PullRequestComment`
         """
         return self.get_review_comments()
 
-    def get_review_comments(self):
+    def get_review_comments(self, since=github.GithubObject.NotSet):
         """
         :calls: `GET /repos/:owner/:repo/pulls/:number/comments <http://developer.github.com/v3/pulls/comments>`_
+        :param since: datetime.datetime format YYYY-MM-DDTHH:MM:SSZ
         :rtype: :class:`github.PaginatedList.PaginatedList` of :class:`github.PullRequestComment.PullRequestComment`
         """
+        assert since is github.GithubObject.NotSet or isinstance(since, datetime.datetime), since
+        url_parameters = dict()
+        if since is not github.GithubObject.NotSet:
+            url_parameters["since"] = since.strftime("%Y-%m-%dT%H:%M:%SZ")
         return github.PaginatedList.PaginatedList(
             github.PullRequestComment.PullRequestComment,
             self._requester,
             self.url + "/comments",
+            url_parameters
+        )
+
+    def get_single_review_comments(self, id):
+        """
+        :calls: `GET /repos/:owner/:repo/pulls/:number/review/:id/comments <https://developer.github.com/v3/pulls/reviews/>`_
+        :param id: integer
+        :rtype: :class:`github.PaginatedList.PaginatedList` of :class:`github.PullRequestComment.PullRequestComment`
+        """
+        assert isinstance(id, (int, long)), id
+        return github.PaginatedList.PaginatedList(
+            github.PullRequestComment.PullRequestComment,
+            self._requester,
+            self.url + "/reviews/" + str(id) + "/comments",
             None
         )
 
@@ -468,7 +607,7 @@ class PullRequest(github.GithubObject.CompletableGithubObject):
         assert isinstance(id, (int, long)), id
         headers, data = self._requester.requestJsonAndCheck(
             "GET",
-            self._parentUrl(self._parentUrl(self.url)) + "/issues/comments/" + str(id)
+            self._parentUrl(self.issue_url) + "/comments/" + str(id)
         )
         return github.IssueComment.IssueComment(self._requester, headers, data, completed=True)
 
@@ -480,10 +619,123 @@ class PullRequest(github.GithubObject.CompletableGithubObject):
         return github.PaginatedList.PaginatedList(
             github.IssueComment.IssueComment,
             self._requester,
-            self._parentUrl(self._parentUrl(self.url)) + "/issues/" + str(self.number) + "/comments",
+            self.issue_url + "/comments",
             None
         )
 
+    def get_review(self, id):
+        """
+        :calls: `GET /repos/:owner/:repo/pulls/:number/reviews/:id <https://developer.github.com/v3/pulls/reviews>`_
+        :param id: integer
+        :rtype: :class:`github.PullRequestReview.PullRequestReview`
+        """
+        assert isinstance(id, (int, long)), id
+        headers, data = self._requester.requestJsonAndCheck(
+            "GET",
+            self.url + "/reviews/" + str(id),
+        )
+        return github.PullRequestReview.PullRequestReview(self._requester, headers, data, completed=True)
+
+    def get_reviews(self):
+        """
+        :calls: `GET /repos/:owner/:repo/pulls/:number/reviews <https://developer.github.com/v3/pulls/reviews/>`_
+        :rtype: :class:`github.PaginatedList.PaginatedList` of :class:`github.PullRequestReview.PullRequestReview`
+        """
+        return github.PaginatedList.PaginatedList(
+            github.PullRequestReview.PullRequestReview,
+            self._requester,
+            self.url + "/reviews",
+            None,
+        )
+
+    def get_review_requests(self):
+        """
+        :calls: `GET /repos/:owner/:repo/pulls/:number/requested_reviewers <https://developer.github.com/v3/pulls/review_requests/>`_
+        :rtype: tuple of :class:`github.PaginatedList.PaginatedList` of :class:`github.NamedUser.NamedUser` and of :class:`github.PaginatedList.PaginatedList` of :class:`github.Team.Team`
+        """
+        return (
+            github.PaginatedList.PaginatedList(
+                github.NamedUser.NamedUser,
+                self._requester,
+                self.url + "/requested_reviewers",
+                None,
+                list_item='users'
+            ),
+            github.PaginatedList.PaginatedList(
+                github.Team.Team,
+                self._requester,
+                self.url + "/requested_reviewers",
+                None,
+                list_item='teams'
+            )
+        )
+
+    def get_labels(self):
+        """
+        :calls: `GET /repos/:owner/:repo/issues/:number/labels <http://developer.github.com/v3/issues/labels>`_
+        :rtype: :class:`github.PaginatedList.PaginatedList` of :class:`github.Label.Label`
+        """
+        return github.PaginatedList.PaginatedList(
+            github.Label.Label,
+            self._requester,
+            self.issue_url + "/labels",
+            None
+        )
+
+    def add_to_labels(self, *labels):
+        """
+        :calls: `POST /repos/:owner/:repo/issues/:number/labels <http://developer.github.com/v3/issues/labels>`_
+        :param label: :class:`github.Label.Label` or string
+        :rtype: None
+        """
+        assert all(isinstance(element, (github.Label.Label, str, unicode)) for element in labels), labels
+        post_parameters = [label.name if isinstance(label, github.Label.Label) else label for label in labels]
+        headers, data = self._requester.requestJsonAndCheck(
+            "POST",
+            self.issue_url + "/labels",
+            input=post_parameters
+        )
+
+    def delete_labels(self):
+        """
+        :calls: `DELETE /repos/:owner/:repo/issues/:number/labels <http://developer.github.com/v3/issues/labels>`_
+        :rtype: None
+        """
+        headers, data = self._requester.requestJsonAndCheck(
+            "DELETE",
+            self.issue_url + "/labels"
+        )
+
+    def remove_from_labels(self, label):
+        """
+        :calls: `DELETE /repos/:owner/:repo/issues/:number/labels/:name <http://developer.github.com/v3/issues/labels>`_
+        :param label: :class:`github.Label.Label` or string
+        :rtype: None
+        """
+        assert isinstance(label, (github.Label.Label, str, unicode)), label
+        if isinstance(label, github.Label.Label):
+            label = label._identity
+        else:
+            label = urllib.quote(label)
+        headers, data = self._requester.requestJsonAndCheck(
+            "DELETE",
+            self.issue_url + "/labels/" + label
+        )
+
+    def set_labels(self, *labels):
+        """
+        :calls: `PUT /repos/:owner/:repo/issues/:number/labels <http://developer.github.com/v3/issues/labels>`_
+        :param labels: list of :class:`github.Label.Label` or strings
+        :rtype: None
+        """
+        assert all(isinstance(element, (github.Label.Label, str, unicode)) for element in labels), labels
+        post_parameters = [label.name if isinstance(label, github.Label.Label) else label for label in labels]
+        headers, data = self._requester.requestJsonAndCheck(
+            "PUT",
+            self.issue_url + "/labels",
+            input=post_parameters
+        )
+
     def is_merged(self):
         """
         :calls: `GET /repos/:owner/:repo/pulls/:number/merge <http://developer.github.com/v3/pulls>`_
@@ -495,16 +747,25 @@ class PullRequest(github.GithubObject.CompletableGithubObject):
         )
         return status == 204
 
-    def merge(self, commit_message=github.GithubObject.NotSet):
+    def merge(self, commit_message=github.GithubObject.NotSet, commit_title=github.GithubObject.NotSet, merge_method=github.GithubObject.NotSet, sha=github.GithubObject.NotSet):
         """
         :calls: `PUT /repos/:owner/:repo/pulls/:number/merge <http://developer.github.com/v3/pulls>`_
         :param commit_message: string
         :rtype: :class:`github.PullRequestMergeStatus.PullRequestMergeStatus`
         """
         assert commit_message is github.GithubObject.NotSet or isinstance(commit_message, (str, unicode)), commit_message
+        assert commit_title is github.GithubObject.NotSet or isinstance(commit_title, (str, unicode)), commit_title
+        assert merge_method is github.GithubObject.NotSet or isinstance(merge_method, (str, unicode)), merge_method
+        assert sha is github.GithubObject.NotSet or isinstance(sha, (str, unicode)), sha
         post_parameters = dict()
         if commit_message is not github.GithubObject.NotSet:
             post_parameters["commit_message"] = commit_message
+        if commit_title is not github.GithubObject.NotSet:
+            post_parameters["commit_title"] = commit_title
+        if merge_method is not github.GithubObject.NotSet:
+            post_parameters["merge_method"] = merge_method
+        if sha is not github.GithubObject.NotSet:
+            post_parameters["sha"] = sha
         headers, data = self._requester.requestJsonAndCheck(
             "PUT",
             self.url + "/merge",
@@ -515,6 +776,7 @@ class PullRequest(github.GithubObject.CompletableGithubObject):
     def _initAttributes(self):
         self._additions = github.GithubObject.NotSet
         self._assignee = github.GithubObject.NotSet
+        self._assignees = github.GithubObject.NotSet
         self._base = github.GithubObject.NotSet
         self._body = github.GithubObject.NotSet
         self._changed_files = github.GithubObject.NotSet
@@ -530,6 +792,7 @@ class PullRequest(github.GithubObject.CompletableGithubObject):
         self._html_url = github.GithubObject.NotSet
         self._id = github.GithubObject.NotSet
         self._issue_url = github.GithubObject.NotSet
+        self._labels = github.GithubObject.NotSet
         self._merge_commit_sha = github.GithubObject.NotSet
         self._mergeable = github.GithubObject.NotSet
         self._mergeable_state = github.GithubObject.NotSet
@@ -553,6 +816,13 @@ class PullRequest(github.GithubObject.CompletableGithubObject):
             self._additions = self._makeIntAttribute(attributes["additions"])
         if "assignee" in attributes:  # pragma no branch
             self._assignee = self._makeClassAttribute(github.NamedUser.NamedUser, attributes["assignee"])
+        if "assignees" in attributes:  # pragma no branch
+            self._assignees = self._makeListOfClassesAttribute(github.NamedUser.NamedUser, attributes["assignees"])
+        elif "assignee" in attributes:
+            if attributes["assignee"] is not None:
+                self._assignees = self._makeListOfClassesAttribute(github.NamedUser.NamedUser, [attributes["assignee"]])
+            else:
+                self._assignees = self._makeListOfClassesAttribute(github.NamedUser.NamedUser, [])
         if "base" in attributes:  # pragma no branch
             self._base = self._makeClassAttribute(github.PullRequestPart.PullRequestPart, attributes["base"])
         if "body" in attributes:  # pragma no branch
@@ -583,6 +853,8 @@ class PullRequest(github.GithubObject.CompletableGithubObject):
             self._id = self._makeIntAttribute(attributes["id"])
         if "issue_url" in attributes:  # pragma no branch
             self._issue_url = self._makeStringAttribute(attributes["issue_url"])
+        if "labels" in attributes:  # pragma no branch
+            self._labels = self._makeListOfClassesAttribute(github.Label.Label, attributes["labels"])
         if "merge_commit_sha" in attributes:  # pragma no branch
             self._merge_commit_sha = self._makeStringAttribute(attributes["merge_commit_sha"])
         if "mergeable" in attributes:  # pragma no branch
